/*	CodeEdit - A programmers editor for Atheos
	Copyright (c) 2002 Andreas Engh-Halstvedt

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Library General Public
	License as published by the Free Software Foundation; either
	version 2 of the License, or (at your option) any later version.
	
	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Library General Public License for more details.
	
	You should have received a copy of the GNU Library General Public
	License along with this library; if not, write to the Free
	Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
	MA 02111-1307, USA
*/
#include "App.h"
#include "EditWin.h"
#include "SettingsDialog.h"
#include "Dialogs.h"

#include <codeview/Format_C.h>
#include <codeview/Format_java.h>

#include <memory>

#include <gui/filerequester.h>
#include <util/messenger.h>
#include <storage/file.h>
#include <storage/directory.h>

App::App() 
: os::Application("application/x-wnd-nikrowd-" APP_NAME),
  options(NULL), loadRequester(NULL), loadRequesterVisible(false)
{
	loadSettings();
}

App::~App(){
}

bool App::OkToQuit(){
	if(windowList.empty()){
		if(!storeSettings())
			cerr << APP_NAME ": Failed to write settings to disk!\n";
		return true;
	}else{
		for(uint a=0;a<windowList.size();++a){
			if(windowList[a]->SafeLock()==0){
				windowList[a]->PostMessage(os::M_QUIT, windowList[a]);
				windowList[a]->Unlock();
			}
		}
		return false;
	}
}

void App::init(int argc, char** argv){
	if(argc==1){
		openWindow();
	}else{
		for(int a=1;a<argc;++a)
			openWindow(argv[a]);
	}
}

void App::openWindow(const char* path){ 
	EditWin *win=new EditWin(this, path);
	windowList.push_back(win);
}

void App::windowClosed(void* win){
	for(uint a=0;a<windowList.size();++a){
		if(windowList[a]==win){
			windowList.erase(&windowList[a]);
			break;
		}
	}
	
	if(windowList.empty())
		PostMessage(os::M_QUIT);
}

void App::setFormatList(const vector<FormatSet>& s){
	formatList=s;
	
	for(uint a=0;a<windowList.size();++a){
		if(windowList[a]->SafeLock()==0){
			windowList[a]->PostMessage(EditWin::ID_FORMAT_CHANGED, windowList[a]);
			windowList[a]->Unlock();
		}
	}
}

bool App::storeSettings(){
	//first put data into message
	keymap.store(&settings);
	
	settings.RemoveName("formatCount");
	settings.RemoveName("formatName");
	settings.RemoveName("formatFilter");
	settings.RemoveName("formatFormat");
	settings.RemoveName("formatUseTab");
	settings.RemoveName("formatTabSize");
	settings.RemoveName("formatUseTab");
	settings.RemoveName("formatAutoindent");
	settings.RemoveName("formatFgColor");
	settings.RemoveName("formatBgColor");
	
	settings.AddInt32("formatCount", (int32)formatList.size());
	for(uint a=0;a<formatList.size();++a){		
		settings.AddString("formatName", formatList[a].name);
		settings.AddString("formatFilter", formatList[a].filter);
		settings.AddString("formatFormat", formatList[a].formatName);
		settings.AddInt32("formatTabSize", formatList[a].tabSize);
		settings.AddBool("formatUseTab", formatList[a].useTab);
		settings.AddBool("formatAutoindendt", formatList[a].autoindent);
		settings.AddColor32("formatBgColor", formatList[a].bgColor);
		settings.AddColor32("formatFgColor", formatList[a].fgColor);
		
		if(formatList[a].format){
			char tmp[20];
			sprintf(tmp, "format%iStyle", a);
			settings.RemoveName(tmp);
			for(uint b=0;b<formatList[a].format->getStyleCount();++b)
				settings.AddColor32(tmp, formatList[a].format->getStyleColor(b));
		}
	}
	
	//then write message to disk
	try{
		os::File file;// "~/config/" APP_NAME "/settings.conf"
	
		os::Directory d0("~/.");
		os::Directory d1;
		
		//open or create directories
		if(d1.SetTo(d0, "config")!=0){
			if(d0.CreateDirectory("config", &d1)!=0)
				return false;
		}
		if(d0.SetTo(d1, APP_NAME)!=0){
			if(d1.CreateDirectory(APP_NAME, &d0)!=0)
				return false;
		}
		
		if(file.SetTo(d0, "settings.conf", O_CREAT | O_WRONLY)!=0)
			return false;
			
		size_t size=settings.GetFlattenedSize();
		auto_ptr<uchar> buffer(new uchar[size]);//This wil be automatically deleted :)
		
		if(settings.Flatten(buffer.get(), size)!=0){
			return false;
		}
		
		size-=file.Write(buffer.get(), size);
		if(file.Flush()!=0)
			return false;
		
		return (size==0);
	}catch(...){
		return false;
	}
}

bool App::loadSettings(){
	try{
		os::File file("~/config/" APP_NAME "/settings.conf");
	
		off_t size=file.GetSize();
		auto_ptr<uchar> buffer(new uchar[size]);//This wil be automatically deleted :)
		
		if(file.Read(buffer.get(), size)!=size){
			loadDefaults();
			return false;
		}
		
		if(settings.Unflatten(buffer.get())!=0){
			loadDefaults();
			return false;
		}
	}catch(...){
		loadDefaults();
		return false;
	}
	
	keymap.load(&settings);
	
	int32 c;
	if(settings.FindInt32("formatCount", &c)==0){
		formatList.resize(c);
		for(int a=0;a<c;++a){
			settings.FindString("formatName", &formatList[a].name, a);
			settings.FindString("formatFilter", &formatList[a].filter, a);
			settings.FindString("formatFormat", &formatList[a].formatName, a);
			if(formatList[a].formatName=="C formater")
				formatList[a].format=new Format_C();
			else if(formatList[a].formatName=="Java formater")
				formatList[a].format=new Format_java();
			settings.FindInt32("formatTabSize", &formatList[a].tabSize, a);
			settings.FindBool("formatUseTab", &formatList[a].useTab, a);
			settings.FindBool("formatAutoindent", &formatList[a].autoindent, a);
			settings.FindColor32("formatBgColor", &formatList[a].bgColor, a);
			settings.FindColor32("formatFgColor", &formatList[a].fgColor, a);
			
			if(formatList[a].format){
				char tmp[20];
				sprintf(tmp, "format%iStyle", a);
				Color32_s c;
				for(uint b=0;b<formatList[a].format->getStyleCount();++b){
					if(settings.FindColor32(tmp, &c, b)==0)
						formatList[a].format->setStyleColor(b, c);
				}					
			}
		}
	}
	
	return true;
}

void App::loadDefaults(){
	keymap.addMapping("4o", EditWin::ID_OPTIONS);
	keymap.addMapping("2n", EditWin::ID_FILE_NEW);
	keymap.addMapping("2s", EditWin::ID_FILE_SAVE);
	keymap.addMapping("3s", EditWin::ID_FILE_SAVE_AS);
	keymap.addMapping("2o", EditWin::ID_FILE_OPEN);
	keymap.addMapping("2q", EditWin::ID_FILE_CLOSE);
	keymap.addMapping("2f", EditWin::ID_FIND);
	keymap.addMapping("0\20\4", EditWin::ID_FIND_NEXT);//F3
	keymap.addMapping("2r", EditWin::ID_REPLACE);
	keymap.addMapping("2g", EditWin::ID_GOTO_LINE);

	formatList.resize(3);
	formatList[0].name="(default)";
	formatList[0].filter="N/A";
	formatList[0].formatName="Null formater";
	formatList[1].name="C/C++";
	formatList[1].filter=".c;.h;.C;.cpp;.cxx";
	formatList[1].format=new Format_C();
	formatList[1].formatName="C formater";
	formatList[2].name="Java";
	formatList[2].filter=".java";
	formatList[2].format=new Format_java();
	formatList[2].formatName="Java formater";
}


void App::HandleMessage(os::Message *m){
	switch(m->GetCode()){
		case ID_OPEN_FILE:
		{
			if(loadRequester){
				loadRequester->Lock();
				if(!loadRequesterVisible)
					loadRequester->Show();
				loadRequesterVisible=true;
				loadRequester->MakeFocus();
				loadRequester->Unlock();
			}else{
				loadRequester =new FileReq(os::FileRequester::LOAD_REQ,
					new os::Messenger(this), "", os::FileRequester::NODE_FILE, 
					true, new os::Message(ID_DO_OPEN_FILE), NULL, true);

				loadRequester->Show();
				loadRequesterVisible=true;
				loadRequester->MakeFocus();
			}
			break;
		}
		case ID_DO_OPEN_FILE:
		{
			loadRequesterVisible=false;
			string s;
			for(int i=0;m->FindString("file/path", &s, i)==0;++i){
				openWindow(s.c_str());
			}
			break;
		}
		case os::M_FILE_REQUESTER_CANCELED:
			loadRequesterVisible=false;
			break;
		case ID_NEW_FILE:
			openWindow();
			break;
		case ID_SHOW_OPTIONS:
        	if(!options){
        	    options=new SettingsDialog(this);
	        	options->Show();
	        }
        	options->MakeFocus();
		    break;
		case ID_OPTIONS_CLOSED:
		    options=NULL;
		    break;
		case ID_WINDOW_CLOSING:
		{
			void *p;
			if(m->FindPointer("source", &p)==0)
				windowClosed(p);
			break;
		}
#ifndef NDEBUG
		default:
			cout << "App: Unhandled Message: " << m->GetCode() << endl;			
			for(int a=0;a<m->CountNames();++a)
				cout << "\t" << m->GetName(a) << endl;
#endif
	}
}


int main(int argc, char** argv){
	App *app=new App();
	
	app->init(argc, argv);
	app->Run();
	
	return 0;
}





















