/*	CodeEdit - A programmers editor for Atheos
	Copyright (c) 2002 Andreas Engh-Halstvedt

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Library General Public
	License as published by the Free Software Foundation; either
	version 2 of the License, or (at your option) any later version.
	
	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Library General Public License for more details.
	
	You should have received a copy of the GNU Library General Public
	License along with this library; if not, write to the Free
	Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
	MA 02111-1307, USA
*/
#include "EditWin.h"
#include "App.h"
#include "StatusBar.h"
#include "Dialogs.h"

#include <codeview/CodeView.h>
#include <codeview/Format_C.h>

#include <gui/menu.h>
#include <gui/textview.h>
#include <gui/desktop.h>
#include <storage/file.h>
#include <util/regexp.h>

#include <fstream>
#include <memory>

uint EditWin::unnamedCounter=0;


EditWin::EditWin(App* a, const char* path)
: os::Window(os::Rect(50, 50, 750, 850), "EditWin", ""),
  changed(false), app(a),
  searchText(""), searchDown(true), searchCase(false), searchExtended(false), 
  findDialog(NULL),
  gotoLineDialog(NULL),
  saveRequester(NULL), saveRequesterVisible(false)
{
	os::Rect r(50, 50, 750, 850);
	os::Desktop desktop;
	os::IPoint dp=desktop.GetResolution();
	if(dp.x<r.right+50)
		r.right=dp.x-50;
	if(dp.y<r.bottom+50)
		r.bottom=dp.y-50;
	SetFrame(r);
	
	r=GetBounds();

	//Make Menus
	menu=new os::Menu(os::Rect(),	"main menu", os::ITEMS_IN_ROW);

	os::Menu *m=new os::Menu(os::Rect(), "Application", os::ITEMS_IN_COLUMN);
	m->AddItem("Quit", new os::Message(ID_QUIT));
	m->AddItem("About...", new os::Message(ID_ABOUT));
	m->AddItem("Options...", new os::Message(ID_OPTIONS));
	menu->AddItem(m);

	m=new os::Menu(os::Rect(), "File", os::ITEMS_IN_COLUMN);
	m->AddItem("New", new os::Message(ID_FILE_NEW));
	m->AddItem("Open...", new os::Message(ID_FILE_OPEN));
	m->AddItem("Save", new os::Message(ID_FILE_SAVE));
	m->AddItem("Save As...", new os::Message(ID_FILE_SAVE_AS));
	m->AddItem("Close", new os::Message(ID_FILE_CLOSE));
	menu->AddItem(m);

	m=new os::Menu(os::Rect(), "Edit", os::ITEMS_IN_COLUMN);
	m->AddItem("Undo", new os::Message(ID_EDIT_UNDO));
	m->AddItem("Redo", new os::Message(ID_EDIT_REDO));
	m->AddItem(new os::MenuSeparator());
	m->AddItem("Cut", new os::Message(ID_EDIT_CUT));
	m->AddItem("Copy", new os::Message(ID_EDIT_COPY));
	m->AddItem("Paste", new os::Message(ID_EDIT_PASTE));
	m->AddItem("Delete", new os::Message(ID_EDIT_DELETE));
	menu->AddItem(m);

	m=new os::Menu(os::Rect(), "Search", os::ITEMS_IN_COLUMN);
	m->AddItem("Find", new os::Message(ID_FIND));
	m->AddItem("Find next", new os::Message(ID_FIND_NEXT));
	m->AddItem("Find previous", new os::Message(ID_FIND_PREVIOUS));
/*	m->AddItem(new os::MenuSeparator());
	m->AddItem("Replace", new os::Message(ID_REPLACE));
	m->AddItem("Replace next", new os::Message(ID_REPLACE_NEXT));
	m->AddItem("Replace all", new os::Message(ID_REPLACE_ALL));
	m->AddItem(new os::MenuSeparator());
*/	m->AddItem("Go to line", new os::Message(ID_GOTO_LINE));
	menu->AddItem(m);

	menu->SetTargetForItems(this);
	
	os::Point p=menu->GetPreferredSize(false);
	menu->SetFrame(os::Rect(r.left, r.top, r.right, r.top+p.y));
	r.top+=p.y+1;
	AddChild(menu);
	
	statusbar=new StatusBar(os::Rect(), "", 3);
	p=statusbar->GetPreferredSize(false);
	statusbar->SetFrame(os::Rect(r.left, r.bottom-p.y, r.right, r.bottom));
	r.bottom-=p.y+1;
	AddChild(statusbar);
	statusbar->configurePanel(0, StatusBar::CONSTANT, 50);
	statusbar->configurePanel(1, StatusBar::CONSTANT, 50);
	statusbar->configurePanel(2, StatusBar::FILL, 0);
	
	edit=new CodeView(r, "", "", os::CF_FOLLOW_ALL);
	edit->SetTarget(this);
	edit->SetMessage(new os::Message(ID_TEXT_INVOKED));
	edit->SetEventMask(os::TextView::EI_CONTENT_CHANGED | os::TextView::EI_CURSOR_MOVED);
	AddChild(edit);
	
	if(path){
		load(path);
		title=path;
		named=true;
	}else{
		uint nr=++unnamedCounter;
		
		char buf[20];
		sprintf(buf, "<unnamed %i>", nr);
		title=buf;
		named=false;
	}		
	
	updateTitle();
	updateStatus();	
	formatChanged();
	Show();
	MakeFocus();
	edit->MakeFocus();
}

EditWin::~EditWin(){
	if(saveRequesterVisible)
		saveRequester->Quit();
	
	os::Message m(App::ID_WINDOW_CLOSING);
	m.AddPointer("source", this);
	
	app->PostMessage(&m);
}

void EditWin::appAbout(){
	MsgDialog *md=new MsgDialog(APP_NAME, 
		APP_NAME " is a text editor intended for programming.\n\n"
		"Share and Enjoy!\n",
		0, "Ok", NULL);
	md->center(this);
	md->addMapping("0\n", 0);
	md->addMapping("0\e", 0);
	md->addMapping("0 ", 0);
	md->addMapping("0o", 0);
	md->Go();
}

void EditWin::appQuit(){
	app->PostMessage(os::M_QUIT);
}

void EditWin::appOptions(){
	app->PostMessage(App::ID_SHOW_OPTIONS);
}

void EditWin::fileNew(){
	app->PostMessage(App::ID_NEW_FILE);
}

void EditWin::fileOpen(){
	app->PostMessage(App::ID_OPEN_FILE);
}

bool EditWin::fileSave(const string &p){
	if(saveRequesterVisible){
		saveRequester->MakeFocus();
		return false;
	}
	
	if(!named && p.empty()){
		fileSaveAs();
		return false;
	}else{
		string path;
		if(p.empty())
			path=title;
		else
			path=p;
			
		try{
			ofstream out(path.c_str());

			for(uint a=0;a<edit->GetLineCount();++a){
				out << edit->GetLine(a) << endl;
			}
		
			title=path;
			changed=false;
			named=true;
			updateTitle();
			formatChanged();
			MakeFocus();
			return true;
		}catch(...){
			string s="Couldn't save to file '";
			s+=path;
			s+="'\n";
			updateStatus(s);
			return false;
		}
	}
}

void EditWin::fileSaveAs(os::Message *parent){
	if(saveRequesterVisible){
		saveRequester->MakeFocus();
	}else{
		os::Message *m=new os::Message(ID_DO_FILE_SAVE_AS);
		if(parent)
			m->AddPointer("Parent", parent);
		
		saveRequester=new FileReq(os::FileRequester::SAVE_REQ,
			new os::Messenger(this), "", 
			os::FileRequester::NODE_FILE, 
			false, m, NULL, true);
	
		//TODO: set path

		saveRequester->Show();
		saveRequesterVisible=true;
		saveRequester->MakeFocus();
	}
}

void EditWin::fileClose(){
	if(changed){
		string s="The file '";
		s+=title;
		s+="' has changed. Do you want to save it?";

		os::Message *m=new os::Message(ID_DO_FILE_SAVE);
		m->AddPointer("Parent", new os::Message(ID_FILE_CLOSE));

		MsgDialog *md=new MsgDialog(APP_NAME, s, 0, "Yes", "No", "Cancel", NULL);
		md->center(this);
		md->addMapping("0y", 0);
		md->addMapping("0\n", 0);
		md->addMapping("0n", 1);
		md->addMapping("0c", 2);
		md->addMapping("0\e", 2);
		md->Go(new os::Invoker(m, this));
	}else{
		Quit();
	}
}

void EditWin::editUndo(){
	edit->Undo();
}

void EditWin::editRedo(){
	edit->Redo();
}

void EditWin::editCut(){
	edit->Cut();
}

void EditWin::editCopy(){
	edit->Copy();
}

void EditWin::editPaste(){
	edit->Paste();
}

void EditWin::editDelete(){
	edit->Delete();
}



void EditWin::updateTitle(){
	string tmp=APP_NAME;
	tmp+=" : ";
	tmp+=title;
	
	if(changed)
		tmp+="*";
		
	SetTitle(tmp);
}

void EditWin::updateStatus(const string &s, bigtime_t timeout){
	char tmp[20];
	os::IPoint p=edit->GetCursor();
	
	sprintf(tmp, "Line %i", p.y+1);
	statusbar->setText(tmp, 0);
	
	sprintf(tmp, "Col %i", p.x);
	statusbar->setText(tmp, 1);
	
	statusbar->setText(s, 2, timeout);
}

bool EditWin::load(const string &path){
	try{		
		os::File f(path);
		off_t size=f.GetSize();
		auto_ptr<char> buf(new char[size+1]);
		f.Read(buf.get(), size);//how about the return value?
		buf.get()[size]=0;
		
		edit->SetEventMask(0);
		edit->Set(buf.get());
		edit->SetEventMask(os::TextView::EI_CONTENT_CHANGED | os::TextView::EI_CURSOR_MOVED);
		title=path;
		named=true;
		changed=false;
		updateTitle();
		formatChanged();
	
		return true;
	}catch(...){
		updateStatus(string("Couldn't open file '") + path + "'!\n");
		return false;
	}		
}

void EditWin::find(const string &str, bool down, bool caseS, bool extended){
	if(str.empty()){
		updateStatus("Search string empty!");
		return;
	}
	
	os::RegExp reg;
	status_t status=reg.Compile(str, !caseS, extended);
	switch(status){
	case os::RegExp::ERR_BADREPEAT:
		updateStatus("Error in regular expression: Bad repeat!");
		return;
	case os::RegExp::ERR_BADBACKREF:
		updateStatus("Error in regular expression: Bad back reference!");
		return;
	case os::RegExp::ERR_BADBRACE:
		updateStatus("Error in regular expression: Bad brace!");
		return;
	case os::RegExp::ERR_BADBRACK:
		updateStatus("Error in regular expression: Bad bracket!");
		return;
	case os::RegExp::ERR_BADPARENTHESIS:
		updateStatus("Error in regular expression: Bad parenthesis!");
		return;
	case os::RegExp::ERR_BADRANGE:
		updateStatus("Error in regular expression: Bad range!");
		return;
	case os::RegExp::ERR_BADSUBREG:
		updateStatus("Error in regular expression: Bad number in \\digit construct!");
		return;
	case os::RegExp::ERR_BADCHARCLASS:
		updateStatus("Error in regular expression: Bad character class name!");
		return;
	case os::RegExp::ERR_BADESCAPE:
		updateStatus("Error in regular expression: Bad escape!");
		return;
	case os::RegExp::ERR_BADPATTERN:
		updateStatus("Error in regular expression: Syntax error!");
		return;
	case os::RegExp::ERR_TOLARGE:
		updateStatus("Error in regular expression: Expression to large!");
		return;
	case os::RegExp::ERR_NOMEM:
		updateStatus("Error in regular expression: Out of memory!");
		return;
	case os::RegExp::ERR_GENERIC:
		updateStatus("Error in regular expression: Unknown error!");
		return;
	}
	
	os::IPoint start=edit->GetCursor();
	
	if(down){
		//test current line
		if(reg.Search(edit->GetLine(start.y), start.x)){//found!
			edit->SetCursor(start.x+reg.GetEnd(), start.y, false, false);
			edit->Select(os::IPoint(start.x+reg.GetStart(), start.y), os::IPoint(start.x+reg.GetEnd(), start.y));
			Flush();
			return;
		}
	
		//test to end of file
		for(uint a=start.y+1;a<edit->GetLineCount();++a){
			if(reg.Search(edit->GetLine(a))){//found!
				edit->SetCursor(reg.GetEnd(), a, false, false);
				edit->Select(os::IPoint(reg.GetStart(), a), os::IPoint(reg.GetEnd(), a));
				Flush();
				return;
			}
		}
		
		//test from start of file
		for(int a=0;a<=start.y;++a){
			if(reg.Search(edit->GetLine(a))){//found!
				edit->SetCursor(reg.GetEnd(), a, false, false);
				edit->Select(os::IPoint(reg.GetStart(), a), os::IPoint(reg.GetEnd(), a));
				Flush();
				return;
			}
		}
		
		//not found
		updateStatus("Search string not found!");		
		
	}else{//search up
		//this isn't nice, but emulates searching from the end of each line
		
		//test current line
		if(reg.Search(edit->GetLine(start.y), 0, start.x)){//found!
			int begin=reg.GetStart();
			int end=reg.GetEnd();
			
			while(reg.Search(edit->GetLine(start.y), begin+1, start.x-begin-1)){
				end=begin+reg.GetEnd()+1;
				begin+=reg.GetStart()+1;
			};
			
			edit->SetCursor(begin, start.y, false, false);
			edit->Select(os::IPoint(begin, start.y), os::IPoint(end, start.y));
			Flush();
			return;
		}
	
		//test to start of file
		for(int a=start.y-1;a>=0;--a){
			if(reg.Search(edit->GetLine(a))){//found!
				int begin=reg.GetStart();
				int end=reg.GetEnd();
			
				while(reg.Search(edit->GetLine(a), begin+1)){
					end=begin+reg.GetEnd()+1;
					begin+=reg.GetStart()+1;
				};
			
				edit->SetCursor(begin, a, false, false);
				edit->Select(os::IPoint(begin, a), os::IPoint(end, a));
				Flush();
				return;
			}
		}
		
		//test from end of file
		for(int a=edit->GetLineCount()-1;a>=start.y;--a){
			if(reg.Search(edit->GetLine(a))){//found!
				int begin=reg.GetStart();
				int end=reg.GetEnd();
			
				while(reg.Search(edit->GetLine(a), begin+1)){
					end=begin+reg.GetEnd()+1;
					begin+=reg.GetStart()+1;
				};
			
				edit->SetCursor(begin, a, false, false);
				edit->Select(os::IPoint(begin, a), os::IPoint(end, a));
				Flush();
				return;
			}
		}
		
		//not found
		updateStatus("Search string not found!");		
	}
}


void EditWin::formatChanged(){
	const vector<FormatSet> &list=app->getFormatList();
	
	if(list.size()==0)
		return;
		
	if(named){
		for(uint a=1;a<list.size();++a){
			int i0=0;
			int i1=list[a].filter.find(';', 0);
			if(i1<0)
				i1=list[a].filter.size();
			
			while(i0<i1){	
				string filter=list[a].filter.substr(i0, i1-i0);
				
				if(title.size()>filter.size()){
					if(filter==title.c_str()+title.size()-filter.size()){
						//match, load format
						edit->SetFormat(list[a].format);
						edit->SetTabSize(list[a].tabSize);
						edit->SetUseTab(list[a].useTab);
						edit->SetTextColor(list[a].fgColor);
						edit->SetBackground(list[a].bgColor);
						updateStatus(string("Using format ")+list[a].name);
						return;
					}
				}
				
				
				i0=i1+1;
				i1=list[a].filter.find(';', i0);
				if(i1<0)
					i1=list[a].filter.size();
			}
			
		}
	}
	
	//no match, load defaults
	edit->SetFormat(list[0].format);
	edit->SetTabSize(list[0].tabSize);
	edit->SetUseTab(list[0].useTab);
	edit->SetTextColor(list[0].fgColor);
	edit->SetBackground(list[0].bgColor);
	updateStatus("Using default format");
}

bool EditWin::OkToQuit(){
	PostMessage(ID_FILE_CLOSE, this);
	return false;
}


void EditWin::DispatchMessage(os::Message *m, os::Handler *h){
	if(m->GetCode()==os::M_KEY_DOWN){
		string str=KeyMap::encodeKey(m);
		int msg;
		if(!str.empty() && app->getKeyMap().findKey(str, &msg)){
			PostMessage(msg, this);
			return;
		}
	}
	Window::DispatchMessage(m, h);
}


void EditWin::HandleMessage(os::Message* msg){
	switch(msg->GetCode()){
		case ID_ABOUT:
			appAbout();
			break;
		case ID_QUIT:
			appQuit();
			break;
		case ID_OPTIONS:
			appOptions();
			break;
		case ID_FILE_NEW:
			fileNew();
			break;
		case ID_FILE_OPEN:
			fileOpen();
			break;
		case ID_FILE_SAVE:
			fileSave();
			break;
		case ID_FILE_SAVE_AS:
			fileSaveAs();
			break;
		case ID_FILE_CLOSE:
			fileClose();
			break;
		case ID_DO_FILE_SAVE_AS:
		{
			saveRequesterVisible=false;
			saveRequester->Quit();
			saveRequester=NULL;
			
			os::Message *parent=NULL;
			msg->FindPointer("Parent", (void**)&parent);
			
			string s;
			assert(msg->FindString("file/path", &s)==0);
			
			fileSave(s);
			if(parent)
				PostMessage(parent, this);
			break;
		}
		case os::M_FILE_REQUESTER_CANCELED:
			saveRequesterVisible=false;
			saveRequester->Quit();
			saveRequester=NULL;
			
			break;			
		case ID_DO_FILE_SAVE:
		{
			int32 which;
			os::Message* parent;		
			assert(msg->FindInt32("which", &which)==0);
			assert(msg->FindPointer("Parent", (void**)&parent)==0);
			
			switch(which){
				case 0://yes, do save
					if(named){
						if(fileSave(title)){
							PostMessage(parent, this);
						}
					}else
						fileSaveAs(parent);
					break;
				case 1://no, don't save
					changed=false;
					PostMessage(parent, this);
					break;
				case 2://cancel, do nothing
					break;
				default:
					assert(false);
			}
			break;
		}
		case ID_EDIT_UNDO:
			editUndo();
			break;
		case ID_EDIT_REDO:
			editRedo();
			break;
		case ID_EDIT_CUT:
			editCut();
			break;
		case ID_EDIT_COPY:
			editCopy();
			break;
		case ID_EDIT_PASTE:
			editPaste();
			break;
		case ID_EDIT_DELETE:
			editDelete();
			break;
		case ID_TEXT_INVOKED:
		{
			int32 i;
			if(msg->FindInt32("events", &i)==0){
				if(!changed && (i&CodeView::EI_CONTENT_CHANGED)){
					changed=true;
					updateTitle();
				}
				if(i&CodeView::EI_CURSOR_MOVED){
					updateStatus();
				}
			}
			break;
		}
		case ID_FIND:
			if(findDialog){
				findDialog->MakeFocus();
			}else{
				//load selection
				string tmp;
				edit->GetRegion(&tmp);
				if(!tmp.empty())
					searchText=tmp;
					
				findDialog=new FindDialog(searchText, searchDown, searchCase, searchExtended);
				findDialog->center(this);
				findDialog->Go(new os::Invoker(new os::Message(ID_DO_FIND), this));
			}
			break;
		case ID_FIND_NEXT:
			if(searchText.empty())
				PostMessage(ID_FIND, this);
			else
				find(searchText, true, searchCase, searchExtended);
			break;
		case ID_FIND_PREVIOUS:
			if(searchText.empty())
				PostMessage(ID_FIND, this);
			else
				find(searchText, false, searchCase, searchExtended);
			break;			
		case ID_GOTO_LINE:
			if(gotoLineDialog){
				gotoLineDialog->MakeFocus();
			}else{
				gotoLineDialog=new GotoLineDialog();
				gotoLineDialog->center(this);
				gotoLineDialog->Go(new os::Invoker(new os::Message(ID_DO_GOTO_LINE), this));
			}
			break;
		case ID_DO_FIND:
		{
			bool b=false;
			if(msg->FindBool("closed", &b)==0 && b){
				findDialog=NULL;
			}else{
				msg->FindString("text", &searchText);
				msg->FindBool("down", &searchDown);
				msg->FindBool("case", &searchCase);
				msg->FindBool("extended", &searchExtended);
				find(searchText, searchDown, searchCase, searchExtended);
			}
			break;
		}		
		case ID_DO_GOTO_LINE:
		{
			bool b=false;
			if(msg->FindBool("closed", &b)==0 && b){
				gotoLineDialog=NULL;
			}else{
				int32 which;
				assert(msg->FindInt32("which", &which)==0);
				edit->SetCursor(0, which-1);
				edit->MakeFocus();
			}
			break;
		}
		case ID_FORMAT_CHANGED:
			formatChanged();
			break;
#ifndef NDEBUG
		default:
		{
			char tmp[50];
			sprintf(tmp, "Unhandled Message: %i", msg->GetCode());
			updateStatus(tmp);
			cout << "EditWin: Unhandled Message: " << msg->GetCode() << endl;			
			for(int a=0;a<msg->CountNames();++a)
				cout << "\t" << msg->GetName(a) << endl;
		}
#endif
	}
}



























